#!/usr/bin/perl -w

# iptables module for dulog
# v0.1 9/3/01 Michael D. Stenner <mstenner@phy.duke.edu>
# made for dulog 0.2.3
# added to CVS by icon
# $Id: iptables.mod.pl,v 1.1 2001/09/06 20:16:53 icon Exp $

# TODO:
# checks to make sure log entry is processed correctly (including funky logs)
# deal with no log entries correctly
# format to 79 columns

use Socket;

##############################################################################
# read options from environment variables (set by dulog)

#                         option name        default
my $pattern     = option('IPTABLES_PATTERN', 'iptables'); # regex that marks a hit
my $debug       = option('DEBUG',            undef); # dulog in debug mode
my $ldebug      = option('IPTABLES_DEBUG',   undef); # just iptables.mod in debug mode
my $quiet       = option("QUIET",            undef); # dulog in quiet mode
my $input       = option("LOGCAT",           undef); # file to read
my $output      = option("LOGREPORT",        undef); # file to write report to
my $weed        = option("WEED",             undef); # are we removing processed logs
my $confdir     = option("CONFDIR",          "../etc"); # where to find config files
my $trojan_file = "$confdir/trojans.list"; # list of trojans and other nasty things

##############################################################################
# process the options

# open the input file, $LOGCAT or read from stdin if LOGCAT isn't defined.
if (defined($input)) {
    open(INPUT, $input) or mlog(0, "cannot open input file $input\n");
    $weed = defined($weed) ? 1 : 0;
} else {
    *INPUT = *STDIN;
    $weed  = 0
}

# open output file, $LOGREPORT or write to STDOUT if LOGREPORT isn't defined.
if (defined($output)) {
    open(OUTPUT, ">$output") or mlog(0, "cannot open output file $output\n");
} else {
    *OUTPUT = *STDOUT;
}

# we'll write unprocessed logs to a replacement $LOGCAT if weeding is on.
if ($weed) {
    open(NEWIN, ">$input.new") or mlog(0, "cannot open tmp file $input.new\n");
}

# set up logging level.  NOTE: logs at level 0 (serious problems) are reported
# even when "QUIET" is on!
my $loglevel = (defined($ldebug) or defined($debug)) ? 2 : 1;
$loglevel = 0 if defined($quiet);
*LOGOUT = *STDOUT; # this is where the logging goes

##############################################################################
# do the magic (main)

# load the trojans list
my $trojans; loadtrojans();

mlog(1, "beginning to read input");
while (defined($line = <INPUT>)) {
    if ($line =~ /$pattern/) {
	$line =~ /^(\S+\s+\S+\s+\S+)\s+(\S+) (.*?)(IN=.*)$/;
	my $time = $1;
	my $host = $2;
	my $type = $3;
	my $data = parse($4);
	
	# syslog-ng can report hosts in a more complicated way :)
	if ($host =~ m{[@/](\S+)}) { $host = $1; }

	# host report  --  report by scanned host
	$hr = $host . ":" . $data->{'DPTNAME'};
	# step counter
	$hostrep{$hr} = exists($hostrep{$hr}) ? $hostrep{$hr} + 1 : 1;

	# offender report  --  report by offending scanner
	$or = $data->{'SRCHOST'} . ' -> ' . $data->{'DPTNAME'};
	# step counter
	$offrep{$or} = exists($offrep{$or}) ? $offrep{$or} + 1 : 1;

    } else {
	# not an iptables log
	print(NEWIN $line) if $weed; # dump to NEWIN only if we're weeding
    }
}

mlog(1, "writing output");
print OUTPUT "::Scanned hosts::\n(Attempts host:port)\n";
foreach $key (sort(keys(%hostrep))) {
    printf(OUTPUT "%6d %s\n", $hostrep{$key}, $key);
}
print OUTPUT "\n::Offenders::\n";
foreach $key (sort(keys(%offrep))) {
    printf(OUTPUT "%6d %s\n", $offrep{$key}, $key);
}

if ($weed) { # if we're weeding, replace the log file
    mlog(1, "replacing input file $input with $input.new");
    rename("$input.new", $input)
	or mlog(0, "cannot move file $input.new to $input");
}

###################################################################
# subroutines

sub mlog {
    # write the log message if it is at or below the current logging level

    # level 0 -- serious problem, should always be reported
    #            (logs are probably not being reported correctly)
    # level 1 -- notification of important events
    # level 2 -- debugging output, useful only when coding this module

    $level = shift;
    print(LOGOUT "DULOG(iptables.mod): ", @_, "\n") if ($level <= $loglevel);
}

sub loadtrojans {
    # read the trojans list from the file
    mlog(1, "loading trojans list");
    open(TROJANS, $trojan_file) or
	mlog(0, "cannot read trojan file $trojan_file");

    my $line;
    while (defined($line = <TROJANS>)) {
	chomp $line;
	my ($proto, $port, $description) = split(/:/, $line);
	$trojans{"$proto:$port"} = $description;
    }
}

sub option {
    # process options passed in by environment variable
    $op      = shift;
    $default = shift;
    return(exists($ENV{$op}) ? $ENV{$op} : $default);
}

sub parse {
    # parse an iptables log
    my $string = shift; # this is the log itself
    my %data; # will dump data into this

    my @array = split(/\s+/, $string); # split on whitespace
    foreach my $item (@array) {
	if ($item =~ /^(\w+)(=?)(.*)/) { # example: "IN=eth0"
	    $data{$1} = $3;
	}
    }

    my $p = lc($data{'PROTO'}); 
    my $port = $data{'DPT'};
    
    # 123.234.63.23 -> blah.foo.bar.com
    # (or remain the ip if it doesn't resolve)
    my @host_a = gethostbyaddr(pack('CCCC',
				    split(/\./, $data{'SRC'})), AF_INET);
    $data{'SRCHOST'} = defined($host_a[0]) ? $host_a[0] : $data{'SRC'};

    my @serv_a = getservbyport($port, $p); # set service name
    if (scalar(@serv_a)) { # did we find anything
	$data{'DPTNAME'} = $serv_a[0] . " ($port/$p)"; # yes? use that.
    } elsif (exists($trojans{$p . ":" . $data{'DPT'}})) { # no? trojan?
	# yes? use that
	$data{'DPTNAME'} = $trojans{$p . ":" . $data{'DPT'}} . " ($port/$p)";
    } else { # still no?  Fine...
	$data{'DPTNAME'} .= $data{'DPT'} . "/$p"; # just report the port number
    }

    return \%data;
}

